#!/usr/bin/env python
# 786

# Aldy source: common.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from builtins import range

import pkg_resources 
import os
import re
import time
import pprint
import logbook


PROTEINS = { # X is stop
	'TTT': 'F', 'CTT': 'L', 'ATT': 'I', 'GTT': 'V', 'TTC': 'F',
	'CTC': 'L', 'ATC': 'I', 'GTC': 'V', 'TTA': 'L', 'CTA': 'L',
	'ATA': 'I', 'GTA': 'V', 'TTG': 'L', 'CTG': 'L', 'ATG': 'M',
	'GTG': 'V', 'TCT': 'S', 'CCT': 'P', 'ACT': 'T', 'GCT': 'A',
	'TCC': 'S', 'CCC': 'P', 'ACC': 'T', 'GCC': 'A', 'TCA': 'S',
	'CCA': 'P', 'ACA': 'T', 'GCA': 'A', 'TCG': 'S', 'CCG': 'P',
	'ACG': 'T', 'GCG': 'A', 'TAT': 'Y', 'CAT': 'H', 'AAT': 'N',
	'GAT': 'D', 'TAC': 'Y', 'CAC': 'H', 'AAC': 'N', 'GAC': 'D',
	'TAA': 'X', 'CAA': 'Q', 'AAA': 'K', 'GAA': 'E', 'TAG': 'X',
	'CAG': 'Q', 'AAG': 'K', 'GAG': 'E', 'TGT': 'C', 'CGT': 'R',
	'AGT': 'S', 'GGT': 'G', 'TGC': 'C', 'CGC': 'R', 'AGC': 'S',
	'GGC': 'G', 'TGA': 'X', 'CGA': 'R', 'AGA': 'R', 'GGA': 'G',
	'TGG': 'W', 'CGG': 'R', 'AGG': 'R', 'GGG': 'G'
}

REV_COMPLEMENT = {
	'A': 'T', 'T': 'A',
	'C': 'G', 'G': 'C',
}


def inspect():
	import code
	code.interact(local=locals())


def sorted_tuple(x):
	return tuple(sorted(x))


def allele_key(x):
	p = re.split(r'(\d+)', x)
	return p[1]


def region_sort_key(x):
	return tuple(int(z) if z.isdigit() else z for z in re.split('(\d+)', x))


def sort_key(x):
	p = re.split(r'(\d+)', x)
	return (int(p[1]),) + tuple(p[2:])


def var_to_allele(x):
	return re.split(r'[\^_]', x)[0]


def rev_comp(seq):
	return ''.join([REV_COMPLEMENT[x] for x in seq[::-1]])


def seq_to_amino(seq):
	return ''.join(PROTEINS[seq[i:i + 3]] for i in range(0, len(seq) - len(seq) % 3, 3))


def check_functional(gene, m):
	if m.pos not in gene.coding_region:
		return False
	if m.op[:3] != 'SNP':
		return True
	seq = ''.join(gene.coding_region[i] if i != m.pos else m.op[5] for i in sorted(gene.coding_region.keys()))
	amino = seq_to_amino(rev_comp(seq) if gene.rev_comp else seq)
	return amino != gene.aminoacid


def static_vars(**kwargs):
	def decorate(func):
		for k in kwargs:
			setattr(func, k, kwargs[k])
		return func
	return decorate


def timing(f):
	def wrap(*args):
		time1 = time.time()
		ret = f(*args)
		time2 = time.time()
		log.debug('Time needed: ({:.1f})', time2 - time1)
		return ret
	return wrap


@static_vars(pp=pprint.PrettyPrinter(indent=4))
def pr(x):
	return pr.pp.pformat(x)


def script_path(path, file):
	return pkg_resources.resource_filename(path, file)


def colorize(text, color='green'):
	return logbook._termcolors.colorize(color, text)


log = logbook.Logger('Aldy')
LOG_FORMAT = '{record.message}'



from __future__ import print_function
# 786

# Aldy source: diplotype.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from builtins import map
from collections import defaultdict, Counter

import sys

from .common import *


def write_decomposition(sol_id, decomposition, genotype, f):
	print('# Solution {}'.format(sol_id), file=f)
	print('# Predicted diplotype: {}'.format(genotype[0]), file=f)
	print('# Composition: {}'.format(','.join(genotype[1])), file=f)
	print('\t'.join('Copy Allele Location Type Coverage Effect dbSNP Code Status'.split()), file=f)
	for i, (allele, mutations) in enumerate(decomposition):
		for t, m, c in sorted(mutations, key=lambda x: x[1].pos):
			print('\t'.join(map(str, [
				i, allele[0] if '/' in allele[0] else allele[1],
				m.pos, m.op, c,
				['NEUTRAL', 'DISRUPTING'][0 if m.functional == 0 else 1],
				m.aux['dbsnp'], m.aux['old'],
				t
			])), file=f)
	print(file=f)


def assign_diplotype(gene, solutions, output):
	if output == '':
		f = sys.stdout
	else:
		f = open(output, 'w')

	print('# Aldy v1.0', file=f)
	print('# Gene: {}'.format(gene.name), file=f)
	print('# Number of solutions: {}'.format(len(solutions)), file=f)
	print(file=f)

	results = []
	for sol_id, (orig_sol, decomposition) in enumerate(solutions):
		# solution is the array of (major, minor) tuples
		sol = [allele_key(x[0]) for x in orig_sol]
		if len(sol) == 1:
			res = '*{}/*{}'.format(gene.deletion_allele, *sol)
		elif len(sol) == 2:
			res = '*{}/*{}'.format(*sol)
		else:
			sol = defaultdict(int, Counter(sol))
			diplotype, dc = [[], []], 0
			# Handle tandems
			for ta, tb in gene.common_tandems:
				while sol[ta] > 0 and sol[tb] > 0:
					diplotype[dc % 2] += [ta, tb]
					dc += 1
					sol[ta] -= 1
					sol[tb] -= 1
			# Handle duplicates
			for allele, count in sol.items():
				if count > 0:
					diplotype[dc % 2] += count * [allele]
					dc += 1
			if len(diplotype[1]) == 0:
				diplotype[1] = [diplotype[0][-1]]
				diplotype[0] = diplotype[0][:-1]
			if len(diplotype[0]) > len(diplotype[1]):
				diplotype[0], diplotype[1] = diplotype[1], diplotype[0]
			res = '/'.join(['+'.join(['*' + y for y in x]) for x in diplotype])

		minor_sol = []
		for al in orig_sol:
			if '/' in al[0]:
				al = '{}|{}'.format(*al)
			elif '+' in al[0]:
				al = al[1] + '+'
			else:
				al = al[1]
			minor_sol.append(al)
		if len(minor_sol) == 1:
			minor_sol.append(gene.deletion_allele)

		results.append((res, minor_sol))
		write_decomposition(sol_id, decomposition, (res, minor_sol), f)

	if output != '':
		f.close()
	return results



from __future__ import division
# 786

# Aldy source: refiner.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from builtins import map
from builtins import str
from builtins import range
from pprint import pprint

import collections
import functools
import multiprocessing

from . import lpinterface

from .common import *
from .gene import Mutation
from .filtering import cnv_filter


def get_refined_solution(gene, sam, initial_solutions, solver):
	mutations = set()
	for sol in initial_solutions: # tuple (solution, cn_config)
		for an in sol[0]:
			mutations |= gene.alleles[an].functional_mutations
			for sa in list(gene.alleles[an].suballeles.values()):
				mutations |= sa.neutral_mutations

	# TODO fix this!
	# Use only relevant mutations
	for p in sam.coverage:
		for o in sam.coverage[p]:
			if o != '_' and Mutation(p, o) not in mutations:
				sam.coverage[p][o] = 0

	# pool = multiprocessing.Pool(4)
	results = list(map(
		functools.partial(refine_ilp, gene=gene, sam=sam, mutations=mutations, solver=solver),
		initial_solutions
	))
	if len(results) > 0:
		score = min(results)[0]
		solutions = [(r[1], r[2]) for r in results if abs(r[0] - score) < 1e-6]
	else:
		score, solutions = float('inf'), []
	return score, solutions


def refine_ilp(solution, gene, sam, mutations, solver):
	log.debug('== Refiner ==')

	solution, region_cn = solution
	log.debug('Refining {}', solution)

	c = lpinterface.model('aldy_refine', solver)
	cnv_filter(sam, gene, region_cn)

	# List all alleles!
	alleles = { # key is (major, minor, copy)
		(a, sa, 0): set(gene.alleles[a].functional_mutations) | set(gene.alleles[a].suballeles[sa].neutral_mutations)
		for a in solution
		for sa in gene.alleles[a].suballeles
	}

	count = collections.Counter(solution)
	for a, sa, _ in list(alleles.keys()):
		for cnt in range(1, count[a]):
			alleles[(a, sa, cnt)] = alleles[(a, sa, 0)]

	# add one binary variable to model for any allele copy
	A = {a: c.addVar(vtype='B', name='{}-{}_{}'.format(*a)) for a in alleles}

	# a[i] >= a[i + 1] for all i
	for a, sa, cnt in alleles:
		if cnt != 0:
			continue
		for i in range(1, count[a]):
			log.trace('M  Sentinel constraint for {}-{}: A_{} <= A_{}', a, sa, i, i - 1)
			c.addConstr(A[(a, sa, i)] <= A[(a, sa, i - 1)])

	for a, cnt in count.items():
		expr = c.quicksum(var for aa, var in A.items() if aa[0] == a)
		log.trace('M  CN constraint for {}: {} == {}', a, cnt, expr)
		c.addConstr(expr == cnt)

	# for any mutation, add error variable to expression
	constraints = {
		m: [0, c.addVar(lb=-c.INF, ub=c.INF, name=str(m))]
		for m in mutations
	}
	# add binary variable for any allele/mutation pair indicating whether it's missing or not
	MM = {
		a: {m: c.addVar(vtype='B', name='MISS-{}-{}-{}_{}'.format(m, *a)) for m in alleles[a]}
		for a in alleles
	}
	# indicating is it added or no
	MA = {
		a: {m: c.addVar(vtype='B', name='ADD-{}-{}-{}_{}'.format(m, *a)) for m in mutations if m not in alleles[a]}
		for a in alleles
	}
	# populate constraints
	for m in mutations:
		coverage = 0
		if region_cn[gene.region_at[m.pos]] != 0:
			coverage = sam.total(m.pos) / region_cn[gene.region_at[m.pos]]

		for a in alleles:
			# print m.pos, gene.region_at[m.pos], gene.cnv_configurations[gene.alleles[a[0]].cnv_configuration].keys()
			if m in alleles[a]:
				constraints[m][0] += coverage * MM[a][m] * A[a]
			elif gene.cnv_configurations[gene.alleles[a[0]].cnv_configuration][gene.region_at[m.pos]] > 0:
				constraints[m][0] += coverage * MA[a][m] * A[a]

		# populate constraints for non-variations
		ref_m = Mutation(pos=m.pos, op='REF')
		if ref_m not in constraints:
			constraints[ref_m] = [0, c.addVar(lb=-c.INF, ub=c.INF, name=str(ref_m))]

		for a in alleles:
			if gene.cnv_configurations[gene.alleles[a[0]].cnv_configuration][gene.region_at[m.pos]] == 0:
				continue

			if m in alleles[a]:
				constraints[ref_m][0] += coverage * (1 - MM[a][m]) * A[a]
			elif not any(m.pos == x.pos for x in alleles[a]):
				constraints[ref_m][0] += coverage * (1 - MA[a][m]) * A[a]

	# make sure that each constraint equals its coverage
	for m, (expr, err) in constraints.items():
		log.trace('M  Contraint for {} {}: {} == {} + err', m, gene.region_at[m.pos], sam[m], expr)
		c.addConstr(expr + err == sam[m])
	# make sure that variation is not assigned to allele if allele does not exist
	for a, mv in MM.items():
		for m, v in mv.items():
			log.trace('M  Contraint for MM[missing] {} {}: {} >= {}', a, m, A[a], v)
			c.addConstr(v <= A[a])
	for a, mv in MA.items():
		for m, v in mv.items():
			log.trace('M  Contraint for MA[extra] {} {}: {} >= {}', a, m, A[a], v)
			c.addConstr(v <= A[a])

	# set objective: minimize absolute sum of errors
	objective = c.abssum([cx[1] for cx in constraints.values()])
	# Non-linear objective linearization for SCIP:
	#            min f(x) <==> min w s.t. f(x) <= w
	w = c.addVar(name='W')
	nonlinear_objective = 0
	for a in alleles:
		nonlinear_objective += 2 * A[a] * sum(1 - v for m, v in MM[a].items())
		objective += sum(v for m, v in MA[a].items())
	c.addConstr(nonlinear_objective <= w)
	objective += w
	log.trace('M  Objective: {}', objective)

	# important rules for mutations:
	# - sub-allele must express ALL its functional mutations!
	for a in alleles:
		p = [MM[a][m] for m in alleles[a] if m.functional]
		if len(p) == 0:
			continue
		log.trace('M  Contraint for {}: {} = {} * A_{}', a, sum(p), len(p), A[a])
		c.addConstr(sum(p) == len(p) * A[a])
	# - if coverage is nil, no allele can select mutation
	for m in (m for a in alleles for m in alleles[a] if sam[m] == 0):
		expr = c.quicksum([MM[a][m] for a in alleles if m in alleles[a]])
		log.trace('M  Contraint for {}: 0 >= {}', m, expr)
		c.addConstr(expr <= 0)
	# - make sure that variation is not over-expressed
	for m in mutations:
		region = gene.region_at[m.pos]
		full_cn = region_cn[region]
		mut_cn = sam.percentage(m)
		if full_cn == 0:
			mut_cn = (0, 0) # upper and lower bound
		else:
			mut_cn /= 100.0 / full_cn
			mut_cn = (int(mut_cn), int(mut_cn + 1))
			if sam[m] > 0 and mut_cn[0] == 0:
				mut_cn = (1, 1)
		expr = c.quicksum([
			MM[a][m] * A[a]
			for a in alleles if m in alleles[a]
		])

		# if it is functional, alleles CANNOT include it via EXTRA
		if not m.functional:
			expr += c.quicksum([
				MA[a][m] * A[a] for a in alleles if m not in alleles[a]
			])
			log.trace('M  Contraint for {}: {} <= {} <= {}', m, mut_cn[0], expr, mut_cn[1])
			c.addConstr(expr >= mut_cn[0])
			c.addConstr(expr <= mut_cn[1])
		else:
			# TODO think about this constraint harder
			# expr += sum([A[a]
			# 	for a in alleles
			# 	if a.split('_')[0].split('$')[0] in gene.fusions_left
			# 	and gene.fusions[a.split('_')[0].split('$')[0]][gene.region_at[m.pos]] > 0])
			# log.debug('HC {} {} {} == {}', region, m, mut_cn, expr)
			log.trace('M  Contraint for {}: {} <= {} <= {}', m, mut_cn[0], expr, mut_cn[1])
			c.addConstr(expr >= mut_cn[0])
			log.trace('M  Contraint for {}: {} <= {}', m, expr, mut_cn[1])
			c.addConstr(expr <= mut_cn[1])

	# solve ILP
	try:
		status, opt = c.solve(objective)
		log.debug('CN Solver status: {}, opt: {}', status, opt)
	except lpinterface.NoSolutionsError:
		return float('inf'), []

	for allele, value in A.items():
		value = int(round(c.getValue(value)))
		if value <= 0:
			continue
		log.debug('  {}: {}', allele, value)
		for m, mv in MM[allele].items():
			mv = c.getValue(mv)
			if round(mv) <= 0:
				continue
			log.debug(
				'    {} {} {:4} ({:3.0f} cp. {:.0f}) {}',
				gene.region_at[m.pos], m,
				sam[m],
				sam.percentage(m),
				sam.percentage(m) / (100.0 / region_cn[gene.region_at[m.pos]]) if region_cn[gene.region_at[m.pos]] != 0 else 0,
				'F' if m.functional else ''
			)
	log.debug('Missing:')
	for allele, value in A.items():
		value = int(round(c.getValue(value)))
		if value <= 0:
			continue
		log.debug('  {}: {}', allele, value),
		for m, mv in MM[allele].items():
			if round(c.getValue(mv)) > 0:
				continue
			log.debug(
				'    {} {} {:4} ({:3.0f}) {}  {:12} {}',
				gene.region_at[m.pos], m,
				sam[m],
				sam.percentage(m),
				'F' if m.functional else '', m.aux['dbsnp'], m.aux['old']
			)
	log.debug('Extra:')
	for allele, value in A.items():
		value = int(round(c.getValue(value)))
		if value <= 0:
			continue
		log.debug('  {}: {}', allele, value),
		for m, mv in MA[allele].items():
			if round(c.getValue(mv)) <= 0:
				continue
			log.debug(
				'    {} {} {:4} ({:3.0f}) {}  {:12} {}',
				gene.region_at[m.pos], m,
				sam[m],
				sam.percentage(m),
				'F' if m.functional else '', m.aux['dbsnp'], m.aux['old']
			)

	decomposition = []
	for allele, value in A.items():
		if int(round(c.getValue(value))) <= 0:
			continue
		mutations = set()
		for m, mv in MM[allele].items():
			mv = c.getValue(mv)
			if int(round(mv)) > 0:
				if 'novel' in m.aux:
					mutations.add(('NOVEL', m, sam[m]))
				else:
					mutations.add(('NORMAL', m, sam[m]))
			else:
				mutations.add(('MISSING', m, sam[m]))
		for m, mv in MA[allele].items():
			if int(round(c.getValue(mv))) > 0:
				mutations.add(('EXTRA', m, sam[m]))
		decomposition.append((allele, mutations))

	solution = sorted([(a[0], a[1]) for a, v in A.items() if int(round(c.getValue(v))) > 0])
	return opt, tuple(solution), decomposition



from __future__ import division
# 786

# Aldy source: protein.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from builtins import map
from builtins import chr
from builtins import str
from builtins import range
from functools import reduce

import collections
import itertools
import functools
import multiprocessing

from . import lpinterface

from .common import *
from .filtering import cnv_filter
from .gene import Allele, Mutation


SHOW_CN_INFO = True


def get_initial_solution(gene, sam, structure, solver):
	# Make sure that the duplicates are gone
	unique_structure = {}
	for i, s in enumerate(structure):
		s, cns = s
		s = collections.Counter(y[0] for y in s if y != gene.deletion_allele)
		key = tuple(sorted(s.elements()))
		if key not in unique_structure:
			unique_structure[key] = (s, cns)
	structure = list(unique_structure.values())

	# pool = multiprocessing.Pool(1)
	results = list(map(
		functools.partial(solve_ilp, gene=gene, sam=sam, solver=solver),
		structure
	))
	if len(results) > 0:
		score, solutions = min(results)[0], results[0][1]
		if len(results) > 1:
			solutions = reduce(lambda x, y: x + y, (r[1] for r in results if abs(r[0] - score) < 1e-6))
	else:
		score, solutions = float('inf'), []

	# if SHOW_CN_INFO:
	# 	log.info(colorize('Copy number best structures:', 'green'))
	# 	res = set()
	# 	for s in best_solutions:
	# 		y = ['1' if i not in gene.fusions else i for i in s[0]]
	# 		y = ['36' if i in ['66', '83', '36', '57'] else i for i in y]
	# 		y = collections.Counter(y)
	# 		y = ', '.join('{}x*{}-like'.format(b, a) for a, b in y.iteritems())
	# 		res.add(y)
	# 	for i, ii in enumerate(res):
	# 		log.info(colorize('  ({}) {}'.format(i + 1, ii), 'green'))

	return score, solutions


def solve_ilp(structure, gene, sam, solver):
	log.debug('== Initial Solver ==')

	structure, region_cn = structure
	log.debug('Solving {}', structure)
	if len(structure) == 0: # only deletions
		return 0, [([gene.deletion_allele] * 2, region_cn)]

	# More filtering
	alleles = []
	cnv_filter(sam, gene, region_cn)
	for an, a in sorted(list(gene.alleles.items()), key=lambda x: sort_key(x[0])):
		remove = any(sam[m] <= 0 for m in a.functional_mutations)
		if remove:
			s = ['Extra Removing {:4}  '.format(an)]
			for m in a.functional_mutations:
				if sam[m] <= 0:
					s.append('{} {}'.format(gene.region_at[m.pos], m))
			log.trace(s[0] + ',  '.join(s[1:]))
		elif a.cnv_configuration in structure:
			alleles.append(an)

	# Check for novel functional mutations
	novel_functional_mutations = set()
	for pos, c in sam.coverage.items():
		for op, cov in c.items():
			if op == '_':
				continue
			if (pos, op) in gene.mutations:
				continue
			# TODO: handle these regions as well
			if gene.region_at[pos][2:] not in gene.unique_regions:
				continue
			if region_cn[gene.region_at[pos]] == 0:
				continue
			# Require AT LEAST 80% coverage per copy
			if sam.percentage(Mutation(pos, op)) < 80.0 / region_cn[gene.region_at[pos]]:
				continue
			if check_functional(gene, Mutation(pos, op)):
				log.debug('Novel mutation: {} {} {} ({} or {}%)', gene.region_at[pos][2:], pos, op, cov, sam.percentage(Mutation(pos, op)))
				novel_functional_mutations.add(Mutation(pos, op, True))
	# log.debug('Novel mutations: {}', list(novel_functional_mutations))

	if set(structure.keys()) - set(gene.alleles[a].cnv_configuration for a in alleles):
		return float('inf'), [] # Some structure has no matching allele; no solutions
	
	rejected = []
	if sam.PHASE:
		for a in alleles:
			muts = sorted((m.pos, m.op) for m in gene.alleles[a].functional_mutations)
			for mi, m in enumerate(muts):
				for mj in range(mi + 1, len(muts)):
					if abs(muts[mi][0] - muts[mj][0]) > 50:
						continue
					if sam.links[(muts[mi], muts[mj])] <= 1:
						# Make sure that we have backup mutations to fall back upon
						# TODO: find better way
						other_mathcing = [s for s in sam.links if s[0] == muts[mi]]
						if len(other_mathcing) > 1:
							rejected.append(a)
							break
				if len(rejected) > 0 and rejected[-1] == a: 
					break
	log.debug('Rejected alleles: {}', rejected)
	alleles = {(a, 0): gene.alleles[a] for a in alleles if a not in rejected}

	c = lpinterface.model('aldy_major_allele', solver)

	log.debug('Possible candidates:')
	for a, _ in sorted(list(alleles.keys()), key=lambda x: sort_key(x[0])):
		log.debug('  {}*{}', gene.name, a)
		for m in sorted(alleles[(a, 0)].functional_mutations, key=lambda m: m.pos):
			log.debug(
				'    {} {} {:4} ({:3.0f}) {} {}',
				gene.region_at[m.pos],
				m, sam[m], sam.percentage(m),
				'F', m.aux['old']
			)
	# create special allele for any possible copy
	for a, _ in list(alleles.keys()):
		max_cn = sum(structure.values())
		log.trace('M  Max. CN for {} = {}', a, max_cn)
		for i in range(1, max_cn):
			alleles[(a, i)] = alleles[(a, 0)]

	# add one binary variable to model for any allele copy
	A = {a: c.addVar(vtype='B', name='A_{}_{}'.format(*a)) for a in alleles}
	# for any mutation, add error variable to expression
	constraints = {
		# m -> constraint, error_var
		m: [0, c.addVar(lb=-c.INF, ub=c.INF, name='MA_{}_{}_{}'.format(m, *a))]
		for a in alleles
		for m in alleles[a].functional_mutations
	}
	# add binary variable for any allele/mutation pair
	M = {
		a: {
			m: c.addVar(vtype='B', name='EXTRA_{}_{}_{}'.format(m, *a))
			for m in constraints
			if m not in alleles[a].functional_mutations
		} for a in alleles
	}
	# populate constraints
	for a in alleles:
		for m in alleles[a].functional_mutations:
			if region_cn[gene.region_at[m.pos]] == 0:
				coverage = 0
			else:
				coverage = max(1, sam.total(m.pos)) / region_cn[gene.region_at[m.pos]]
			constraints[m][0] += coverage * A[a]
	for a in M:
		for m in M[a]:
			if region_cn[gene.region_at[m.pos]] != 0:
				constraints[m][0] += coverage * A[a] * M[a][m]
	# populate constraints for non-variations
	for m in list(constraints.keys()):
		if m.op[:3] == 'INS':
			continue # no need for insertions...
		ref_m = Mutation(pos=m.pos, op='REF')
		if ref_m in constraints:
			ref_m = Mutation(pos=m.pos, op=ref_m.op + '#')
		if ref_m not in constraints:
			constraints[ref_m] = [0, c.addVar(lb=-c.INF, ub=c.INF, name=str(ref_m))]

		if region_cn[gene.region_at[m.pos]] == 0:
			coverage = 0
		else:
			coverage = max(1, sam.total(m.pos)) / region_cn[gene.region_at[m.pos]]
		for a in alleles:
			constraints[ref_m][0] += coverage * A[a]

	# make sure that each constraint equals its coverage
	for m, (expr, err) in constraints.items():
		log.trace('M  Contraint for {} {}: {} == {} + err', m, gene.region_at[m.pos], sam[m], expr)
		c.addConstr(expr + err == sam[m])

	# force alleles to be as much as needed
	for a, cnt in structure.items():
		expr = sum(A[y] for y in A if alleles[y].cnv_configuration == a)
		log.trace('M  CN contraint for {}: {} == {}', a, cnt, expr)
		c.addConstr(expr == cnt)

	# Make sure that A[i+1] <= A[i] (to avoid equivalent solutions)
	for a, ai in alleles:
		if ai != 0:
			continue
		for i in itertools.count(1):
			if (a, i) not in A:
				break
			log.trace('M  Sentinel contraint for {}: A_{} <= A_{}', a, i, i - 1)
			c.addConstr(A[(a, i)] <= A[(a, i - 1)])

	# allele must express all of its functional functional_mutations
	for m in (m for a in alleles for m in alleles[a].functional_mutations if sam[m] > 0):
		expr = c.quicksum([A[a] for a in alleles if m in alleles[a].functional_mutations])
		expr += c.quicksum([A[a] * M[a][m] for a in alleles if m not in alleles[a].functional_mutations])
		log.trace('M  Contraint for {}: 1 <= {}', m, expr)
		c.addConstr(expr >= 1)

	# set objective: minimize absolute sum of errors
	# also make sure that every non-expressed snp gets heavily penalized
	penal = 100000
	objective = c.abssum([cx[1] for cx in constraints.values()]) + penal * c.quicksum([M[a][m] for a in M for m in M[a]])
	log.trace('M  Objective: {}', objective)

	# solve ILP
	try:
		status, opt, solutions = c.solveAll(objective, dict(list(A.items()) + [((a, m), M[a][m]) for a in M for m in M[a]]))
		log.debug('CN Solver status: {}, opt: {}', status, opt)
	except lpinterface.NoSolutionsError:
		return float('inf'), []

	unique_key = collections.defaultdict(str)
	for si, s in enumerate(solutions):
		sd = {x: [y[1] for y in s if isinstance(y[0], tuple) and y[0][0] == x[0]] for x in s if not isinstance(x[0], tuple)}
		solutions[si] = []
		for (a, _), extra in sd.items():
			if len(extra) > 0:
				log.warn('Novel major star-allele (*{}-like) found!', a)
				extra = [Mutation(m.pos, m.op, m.functional, dict(list(m.aux.items()) + [('novel', True)])) for m in extra]
				an = Allele(
					gene.alleles[a].name + '+{}'.format(unique_key[gene.alleles[a].name]),
					gene.alleles[a].cnv_configuration,
					gene.alleles[a].functional_mutations | set(extra),
					gene.alleles[a].suballeles
				)
				unique_key[gene.alleles[a].name] = 'a' if unique_key[gene.alleles[a].name] == '' else chr(ord(unique_key[gene.alleles[a].name]) + 1)
				gene.alleles[an.name] = an
				solutions[si].append(an.name)
			else:
				solutions[si].append(a)
	log.debug('  solution:')
	for s in solutions:
		log.debug('    {}', s)

	return opt, [(s, region_cn) for s in solutions]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 786

# Aldy source: sam.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from __future__ import print_function
from __future__ import division
from future import standard_library
standard_library.install_aliases()
from builtins import range # noqa
from builtins import object # noqa
from pprint import pprint # noqa

import pysam # noqa
import os # noqa
import copy # noqa
import subprocess # noqa
import tempfile # noqa
import collections # noqa
import pickle as pickle # noqa

from .common import * # noqa
from .gene import Mutation # noqa


class SAM(object):
	CACHE = False
	PHASE = False
	PROFILE = False

	def get_cache(self, path, gene):
		return path + '-{}.aldycache'.format(gene.name)

	def __init__(self, sam_path, gene, threshold):
		if SAM.PROFILE:
			self.load_profile(sam_path, threshold)
			exit(0)

		self.coverage = dict()
		self.cnv_coverage = collections.defaultdict(int)
		self.region_coverage = dict()
		self.sequence = ''
		self.unfiltered_coverage = collections.defaultdict(int)
		
		if SAM.CACHE and os.path.exists(self.get_cache(sam_path, gene)):
			log.debug('Loading cache')
			d = pickle.load(open(self.get_cache(sam_path, gene)))
			self.__dict__.update(d)
		else:
			self.load(sam_path, gene)
			if SAM.CACHE:
				pickle.dump(self.__dict__, open(self.get_cache(sam_path, gene), 'wb'))

		avg_coverage = sum(self.total(pos) for pos in self.coverage) / float(len(self.coverage))
		log.debug('Coverage is {}', avg_coverage)
		self.threshold = threshold

	def total(self, pos):
		return float(sum(v for p, v in self[pos].items() if p[:3] != 'INS'))

	def percentage(self, m):
		total = self.total(m.pos)
		if total == 0:
			return 0
		return 100.0 * float(self[m]) / total

	def __getitem__(self, index):
		if isinstance(index, Mutation):
			if index.pos not in self.coverage:
				self.coverage[index.pos] = {}
			op = index.op
			if op[:3] == 'REF':
				op = '_'
			if op not in self.coverage[index.pos]:
				self.coverage[index.pos][op] = 0
			return self.coverage[index.pos][op]
		else:
			if index not in self.coverage:
				self.coverage[index] = {}
			return self.coverage[index]

	def __setitem__(self, index, value):
		if isinstance(index, Mutation):
			self.coverage[index.pos][index.op] = value
		else:
			self.coverage[index] = value

	def dump(self, gene, limit=-1, full=False):
		for pos in sorted(self.coverage):
			if limit != -1 and pos > limit:
				break
			for op, cov in self.coverage[pos].items():
				m = Mutation(pos=pos, op=op)
				if self.percentage(m) < 5:
					continue
				if (full or op != '_') and gene.region_at[pos] != '':
					log.debug(
						' {:6} {:5}: {:10} {:10} {:5} ({:3.0f}%)',
						gene.region_at[pos], pos, op,
						gene.old_notation[m],
						cov, self.percentage(m)
					)

	def load(self, sam_path, gene):
		log.debug('SAM file: {}', sam_path)

		muts = collections.defaultdict(int)
		norm = collections.defaultdict(int)

		self.links = collections.defaultdict(int)

		is_deez_file = sam_path[-3:] == '.dz'
		pipe = ''
		if is_deez_file:
			log.debug('Using DeeZ file {}', sam_path)

			def is_exe(path): # http://stackoverflow.com/questions/377017/test-if-executable-exists-in-python/377028#377028
				return os.path.isfile(path) and os.access(path, os.X_OK)
			if not is_exe('deez'):
				found = False
				for path in os.environ["PATH"].split(os.pathsep):
					path = path.strip('"')
					if is_exe(os.path.join(path, 'deez')):
						found = True
						break
				if not found:
					log.error('DeeZ not in PATH')
					exit(1)

			tmpdir = tempfile.mkdtemp()
			pipe = os.path.join(tmpdir, 'dzpipe')
			os.mkfifo(pipe)
			chromosome, chr_start, chr_end = gene.region
			cnv_chromosome, cnv_start, cnv_end = gene.cnv_region
			cnv_region = '{}:{}-{}'.format(cnv_chromosome, cnv_start - 500, cnv_end + 1)[3:]
			region = '{}:{}-{}'.format(chromosome, chr_start - 500, chr_end + 1)[3:]

			command = "deez --stats {} 2>&1 | grep 'Block' | awk '{{print $3}}'".format(sam_path)
			p = subprocess.check_output(command, shell=True)
			if all(x[:3] == 'chr' or x[:2] == '*:' for x in p.split()):
				cnv_region = 'chr' + cnv_region
				region = 'chr' + region

			command = 'deez {} -h -c -Q -r /cs/compbio3/ibrahim/mpeg/ref/hs37d5.fa "{};{}" > {} 2>/dev/null'.format(
				sam_path, region, cnv_region, pipe)
			log.debug(command)
			p = subprocess.Popen(command, shell=True)
			sam_path = pipe

		ref = gene.seq

		# TODO: improve over non-functional mutations
		self.indel_sites = {}
		for a in gene.alleles:
			for m in gene.alleles[a].functional_mutations:
				if m.op[:3] == 'INS':
					self.indel_sites[m] = [collections.defaultdict(int), 0]

		total_reads = 0
		filtered_reads = 0
		with pysam.AlignmentFile(sam_path) as sam:
			try:
				sam.check_index()
			except AttributeError:
				pass # SAM files do not have index
			except ValueError as be:
				raise be

			region = ''
			chromosome, chr_start, chr_end = gene.region
			cnv_chromosome, cnv_start, cnv_end = gene.cnv_region
			if not chromosome.startswith('chr'):
				chromosome = 'chr' + chromosome
				cnv_chromosome = 'chr' + cnv_chromosome

			# TODO: saner way 
			# if sum(x['SN'][:3] == 'chr' for x in sam.header['SQ']) < 2.0/3.0 * len(sam.header['SQ']):
			chrs = [x['SN'] for x in sam.header['SQ']]
			if 'chr1' not in chrs and '1'in chrs:
				chromosome = chromosome[3:]
				cnv_chromosome = cnv_chromosome[3:]
			
			fetched_cnv = False
			if sam.has_index():
				log.debug('File {} has index', sam_path)
				
				region = '{}:{}-{}'.format(cnv_chromosome, cnv_start - 500, cnv_end + 1)
				for read in sam.fetch(region=region):
					start = read.reference_start
					if read.cigartuples is None or (read.flag & 0x40) == 0:
						continue
					for op, size in read.cigartuples:
						if op in [0, 7, 8, 2]:
							for i in range(size):
								self.cnv_coverage[start + i] += 1
							start += size
				fetched_cnv = True

				region = '{}:{}-{}'.format(chromosome, chr_start - 500, chr_end + 1)

			mutation_sites = set()
			for read in sam.fetch(region=region):
				start = read.reference_start
				if not fetched_cnv and read.reference_id != -1 and read.reference_name == cnv_chromosome \
					and cnv_start - 500 <= start <= cnv_end:
					if read.cigartuples is None or (read.flag & 0x40) == 0:
						pass
					else:
						for op, size in read.cigartuples:
							if op in [0, 7, 8, 2]:
								for i in range(size):
									self.cnv_coverage[start + i] += 1
								start += size

				start, s_start = read.reference_start, 0
				if not read.cigartuples or read.reference_name != chromosome or not (chr_start <= start < chr_end):
					continue

				if 'S' in read.cigarstring:
					sk = 0
					for op, size in read.cigartuples:
						if op == 4:
							sk += size
					filtered_reads += 1
				total_reads += 1

				seq = read.query_sequence
				
				has_indel = set()
				orig_start = start

				paired_muts = []
				for op, size in read.cigartuples:
					if op == 2:
						mut = (start, 'DEL.{}'.format(ref[start - chr_start:start - chr_start + size]))
						muts[mut] += 1
						if SAM.PHASE: 
							mutation_sites.add(start)
							paired_muts.append(mut)
						start += size
					elif op == 1:
						mut = (start, 'INS.{}'.format(seq[s_start:s_start + size].lower()))
						muts[mut] += 1
						if SAM.PHASE:
							mutation_sites.add(start)
							paired_muts.append(mut)
						has_indel.add((start, size))
						s_start += size
					elif op == 4:
						s_start += size
					elif op in [0, 7, 8]:
						for i in range(size):
							if 0 <= start + i - chr_start < len(ref) and ref[start + i - chr_start] != seq[s_start + i]:
								mut = (start + i, 'SNP.{}{}'.format(ref[start + i - chr_start], seq[s_start + i]))
								muts[mut] += 1
								if SAM.PHASE: 
									mutation_sites.add(start + i)
									paired_muts.append(mut)
							else:
								norm[start + i] += 1
								if SAM.PHASE and start + i in mutation_sites:
									paired_muts.append((start + i, '_'))
						start += size
						s_start += size

				# TODO: check is real insertion? (does it match reference/read)
				for ins in self.indel_sites:
					ins_len = len(ins.op) - 4
					if (ins.pos, ins_len) in has_indel:
						self.indel_sites[ins][1] += 1
					else:
						self.indel_sites[ins][0][(orig_start, start, len(read.query_sequence))] += 1

				if SAM.PHASE:
					paired_muts = sorted(paired_muts)
					for pi, p in enumerate(paired_muts):
						for pj in range(pi + 1, len(paired_muts)):
							if p[1] == '_' and paired_muts[pj][1] == '_':
								continue
							self.links[(p, paired_muts[pj])] += 1

		for i, c in norm.items():
			if c == 0:
				continue
			if i not in self.coverage:
				self.coverage[i] = {}
			self.coverage[i]['_'] = c
		for k, c in muts.items():
			p, o = k
			if p not in self.coverage:
				self.coverage[p] = {}
			self.coverage[p][o] = c

		for pos in self.coverage:
			self.unfiltered_coverage[pos] = self.total(pos)
		self.original_coverage = copy.deepcopy(self.coverage)

		if pipe != '' and os.path.exists(pipe):
			os.unlink(pipe)

		# pprint(self.links)

	# PGRNseq-v1: PGXT104 for all except CYP2B4; PGXT147 with rescale 2.52444127771 for CYP2B6
	#    cypiripi.py --generate-profile bams/PGXT149.bam --profile-factor 2.52444127771 | grep CYP2B6 > _ha &&
	#    cypiripi.py --generate-profile bams/PGXT104.bam | grep -v CYP2B6 >> _ha
	# PGRNseq-v2: NA19789 for all
	# Illumina: all ones
	def load_profile(self, sam_path, factor):
		log.debug('SAM file: {}', sam_path)

		R = [
			('CYP3A5', '7', 99245000, 99278000),
			('CYP3A4', '7', 99354000, 99465000),
			('CYP2C19', '10', 96445000, 96615000),
			('CYP2C9', '10', 96691000, 96754000),
			('CYP2C8', '10', 96796000, 96830000),
			('CYP4F2', '19', 15619000, 16009500),
			('CYP2A6', '19', 41347500, 41400000),
			('CYP2D6', '22', 42518900, 42553000),
			('TPMT', '6', 18126541, 18157374),
			('DPYD', '1', 97541298, 98388615)
		#	('CYP2B6', '19', 41428000, 41525000),
		#	('SLCO1B1', '12', 21282127, 21394730),
		# 	('CYP21A', '6', 31970000, 32010000),
		# 	('CLN3', '16', 28488000, 28504000),
		]
		for r in R:
			cov = collections.defaultdict(lambda: collections.defaultdict(int))
			rep_cov = collections.defaultdict(lambda: collections.defaultdict(int))

			gene, r = r[0], (r[1], r[2], r[3])
			with pysam.AlignmentFile(sam_path) as sam:
				chrs = [x['SN'] for x in sam.header['SQ']]
				if r[0] not in chrs and 'chr' + r[0] in chrs:
					r = ('chr' + r[0], r[1], r[2])
				log.info('Generating profile for {} ({}:{}-{})', gene, *r)
				try:
					for read in sam.fetch(region='{}:{}-{}'.format(r[0], r[1], r[2])):
						start, s_start = read.reference_start, 0
						if not read.cigartuples:
							continue
						chr = read.reference_name
						if chr.startswith('chr'):
							chr = chr[3:]
						
						for op, size in read.cigartuples:
							if op == 2:
								for i in range(size):
									cov[chr][start + i] += 1
									if read.mapping_quality == 0:
										rep_cov[chr][start + i] += 1
								start += size
							elif op == 1:
								s_start += size
							elif op == 4:
								s_start += size
							elif op in [0, 7, 8]:
								for i in range(size):
									cov[chr][start + i] += 1
									if read.mapping_quality == 0:
										rep_cov[chr][start + i] += 1
								start += size
								s_start += size
				except ValueError as e:
					log.warn('Cannot fetch gene {} ({}:{}-{})', gene, *r)
					
			for c in sorted(cov.keys()):
				for p in sorted(cov[c].keys()):
					if not r[1] - 500 <= p <= r[2] + 500:
						continue
					print(gene, c, p, cov[c][p] * (factor / 2.0))



#!/usr/bin/env python
# 786

# Aldy source: __main__.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from __future__ import division

import logbook
import argparse
import os

from . import genotype
from . import protein
from . import gene
from . import sam
from . import lpinterface

from .common import *


def get_args():
	parser = argparse.ArgumentParser()
	parser.add_argument('--threshold', '-T', default=50, 
		help="Cut-off rate for variations (percent per copy). Default is 50%%.")
	parser.add_argument('--profile', '-p', default=None, 
		help='Sequencing profile. Currently, only "illumina", "pgrnseq-v1" ' + 
			 'and "pgrnseq-v2" are supported. Please check --generate-profile ' +
			 'for more information how to use your own profile.')
	parser.add_argument('--gene', '-g', default='cyp2d6', 
		help='Gene profile. Default is "CYP2D6".')
	parser.add_argument('--verbosity', '-v', default='INFO', 
		help='Logging verbosity. Acceptable values are ' + 
			 'T (trace), D (debug), I (info) and W (warn). Default is I (info).')
	parser.add_argument('--log', '-l', default=None, 
		help='Location of the output log file (default: [input].[gene].aldylog)')
	parser.add_argument('--output', '-o', default=None, 
		help='Location of the output file (default: [input].[gene].aldy)')
	parser.add_argument('--solver', '-s', default='any', 
		help='IP Solver (default: any available)')
	parser.add_argument('--phase', '-P', default=0, action='store_true', 
		help='Phase reads (default: no; slows down pipeline)')
	parser.add_argument('--license', default=0, action='store_true', 
		help='Show the Aldy license')
	parser.add_argument('--test', default=0, action='store_true', 
		help='Sanity-check Aldy on NA10860 sample')
	parser.add_argument('file', nargs='?', 
		help='SAM or BAM input file')

	parser.add_argument('--show-cn', dest='show_cn', action='store_true',
		help='Show all copy number configurations supported by a gene (requires -g).')
	parser.add_argument('--cn', '-c', default=None,
		help='Manually set copy number (input: a comma-separated list CN1,CN2,...). ' +
			 'For a list of supported configurations, please run --show-cn.')
	parser.add_argument('--generate-profile', dest='generate_profile', action='store_true', 
		help='DO NOT USE / ONLY FOR DEVELOPMENT USE')

	# internal parameters for development purposes: please do not use unless instructed
	parser.add_argument('--cache', '-C', dest='cache', action='store_true', 
		help='DO NOT USE / ONLY FOR DEVELOPMENT USE')
	parser.add_argument('--profile-factor', dest='profile_factor', default=2.0, 
		help='DO NOT USE / ONLY FOR DEVELOPMENT USE')

	return parser.parse_args()


def print_licence():
	with open(script_path('aldy.resources', 'LICENSE.md')) as f:
		for l in f: print(l.strip())


def run_test():
	log.warn('Aldy Sanity-Check Test')
	log.warn('Expected result is: *1/*4+*4\n')
	result = genotype.genotype(
		script_path('aldy.resources', 'NA10860.bam'), 
		'', os.devnull, 'CYP2D6', 'illumina',
		0.5, 'any', None
	)
	log.warn('Result{}: ', '' if len(result) == 1 else 's')
	for rd, r in result:
		log.warn('  {:30} ({})', rd, ', '.join(r))


def main(args=None):
	args = get_args()

	level = args.verbosity.lower()
	for k, v in logbook.base._reverse_level_names.items():
		if k.lower().startswith(level):
			level = v
			break

	sh = logbook.more.ColorizedStderrHandler(format_string=LOG_FORMAT, level=level)
	sh.push_application()

	if args.license:
		print_licence()
		exit(0)
	elif args.test:
		run_test()
		exit(0)
	elif args.show_cn:
		database_file = script_path('aldy.resources.genes', '{}.yml'.format(args.gene.lower()))
		g = gene.Gene(database_file)
		g.print_configurations()
		exit(0)
	elif args.file is None:
		log.critical('Input file must be specified!')
		exit(1)

	if args.generate_profile:
		sam.SAM.PROFILE = True
		sam.SAM(args.file, None, float(args.profile_factor))

	if args.cache:
		sam.SAM.CACHE = True
	if args.phase:
		sam.SAM.PHASE = True

	try:
		m = lpinterface.model('init', args.solver)
	except Exception as e:
		log.critical('{}', e.message)


	if args.profile is None:
		log.critical('No profile provided! Please run aldy --help for available profile choices.')
		exit(1)

	result = genotype.genotype(
		args.file, args.output,
		args.log,
		args.gene, args.profile,
		float(args.threshold) / 100.0,
		args.solver,
		args.cn
	)
	log.warn('Result{}: ', '' if len(result) == 1 else 's')
	for rd, r in result:
		log.warn('  {:30} ({})', rd, ', '.join(r))


if __name__ == "__main__":
	main()



from __future__ import division
# 786

# Aldy source: filtering.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


import copy

from .common import *
from .gene import *


def basic_filter(sam):
	for pos in sam.coverage:
		s = sam.total(pos)
		sam.unfiltered_coverage[pos] = s

		indels = [i for i in sam[pos] if i[:3] == 'INS']
		# Make sure not to filter out large indels
		for i in indels:
			if Mutation(pos, i) in sam.indel_sites:
				current_ratio = float(sam[pos][i]) / s
				j = sam.indel_sites[Mutation(pos, i)]

				# calculate j
				j0 = 0
				for orig_start, start, read_len in j[0]:
					ins_len = len(i) - 4
					# This is rather hackish way of detecting *40
					# min 20% sa strane, max ... how much?
					min_margin = .20 * max(1, ins_len / 10.0) * read_len
					if start < pos + max(3 * ins_len, min_margin) or \
						orig_start > pos - max(2 * ins_len, min_margin):
						continue
					j0 += j[0][(orig_start, start, read_len)]
				j[0] = j0

				new_ratio = j[1] / float(j[0] + j[1])
				log.debug(
					'Refining indel {}:{}: from {} of {} to {} (indel:total = {}:{})',
					pos, i,
					sam[pos][i], s,
					int(sam[pos][i] * (new_ratio / current_ratio)),
					j[1], j[0] + j[1]
				)
				sam[pos][i] = int(sam[pos][i] * (new_ratio / current_ratio))
				sam.original_coverage[pos][i] = sam[pos][i]

		threshold = sam.threshold / 20.0 # MAX copy number
		sam.coverage[pos] = {
			k: v
			for k, v in sam.coverage[pos].items()
			if v >= s * threshold and v > 1
		}


def cnv_filter(sam, gene, copy_number):
	original_coverage = copy.deepcopy(sam.original_coverage)
	sam.coverage = dict()
	for pos in original_coverage:
		s = sam.unfiltered_coverage[pos]
		if gene.region_at[pos] in copy_number and copy_number[gene.region_at[pos]] > 0: # we don't care about other mutations
			s /= copy_number[gene.region_at[pos]]
		sam.coverage[pos] = {
			k: v
			for k, v in original_coverage[pos].items()
			if k == '_' or (v > 1 and v >= s * sam.threshold)
		}


def initial_filter(gene, sam):
	log.debug('== Allele Filter ==')

	basic_filter(sam)
	sam.dump(gene)

	useless = []
	for an, a in sorted(list(gene.alleles.items()), key=lambda x: sort_key(x[0])):
		remove = any(sam[m] <= 0 for m in a.functional_mutations)
		if remove:
			s = ['Removing {:4}  '.format(an)]
			for m in a.functional_mutations:
				if sam[m] <= 0:
					s.append('{} {}'.format(gene.region_at[m.pos], m))
			log.trace(s[0] + ',  '.join(s[1:]))
			useless.append(an)

	for u in useless:
		del gene.alleles[u]

	log.debug('Initial status')
	for an, a in sorted(list(gene.alleles.items()), key=lambda x: sort_key(x[0])):
		if '/' in an:
			continue
		log.debug('  {}*{}', gene.name, an)
		for m in sorted(a.functional_mutations, key=lambda x: x.pos):
			log.debug(
				'    {:14} {:15} {:4} ({:3.0f}) {} {}',
				gene.region_at[m.pos],
				str(m), sam[m], sam.percentage(m),
				'F', m.aux['old']
			)

	return gene.alleles



# 786

# Aldy source: genotype.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from builtins import str

import os
import sys
import traceback
import logbook
import logbook.more
import platform

from . import cn
from . import filtering
from . import protein
from . import refiner
from . import sam
from . import diplotype

from .common import *
from .gene import Gene
from .version import __version__


@timing
def genotype(sample, output, log_output, gene, profile, threshold, solver, cn_solution):
	try:
		with open(sample): # Check does file exist
			pass

		if output is None:
			output = '{}.{}.aldy'.format(os.path.splitext(sample)[0], gene.upper())
		if log_output is None:
			log_output = '{}.{}.aldylog'.format(os.path.splitext(sample)[0], gene.upper())
		fh = logbook.FileHandler(log_output, format_string=LOG_FORMAT, mode='w', bubble=True, level='TRACE')
		fh.push_application()

		log.info('*** Aldy v{} (Python {}) ***', __version__, platform.python_version())
		log.info('(c) 2017 SFU, MIT & IUB. All rights reserved.')
		log.info('Arguments:')
		log.info('  Gene:      {}', gene.upper())
		log.info('  Profile:   {}', profile)
		log.info('  Threshold: {:.0f}%', threshold * 100)
		log.info('  Input:     {}', sample)
		log.info('  Output:    {}', output)
		log.info('  Log:       {}', log_output)
		log.info('  Phasing:   {}', sam.SAM.PHASE)

		database_file = script_path('aldy.resources.genes', '{}.yml'.format(gene.lower()))
		gene = Gene(database_file)

		sample = sam.SAM(sample, gene, threshold)
		gene.alleles = filtering.initial_filter(gene, sample)
		cn_sol = cn.estimate_cn(gene, sample, profile, cn_solution, solver)
		score, init_sol = protein.get_initial_solution(gene, sample, cn_sol, solver)
		score, sol = refiner.get_refined_solution(gene, sample, init_sol, solver)

		sol = diplotype.assign_diplotype(gene, sol, output)

		return sol
	except ValueError as ex:
		log.critical('Input BAM has no index. Please create index by running samtools index.')
		log.debug(ex)
		exit(1)
	except IOError as ex:
		if ex.filename is not None:
			log.critical('File cannot be accessed: {}', ex.filename)
		else:
			log.critical('File cannot be accessed: {}', str(ex))
		log.debug(ex)
		exit(1)
	except SystemExit as ex:
		log.debug(ex)
		exit(ex.code)
	except Exception as ex:
		log.error(ex.message)
		log.debug(traceback.format_exc())
		exit(1)
	except:
		ex = sys.exc_info()[0]
		log.critical('Unrecoverable error: {}', str(ex))
		log.debug(traceback.format_exc())
		exit(1)



#!/usr/bin/env python
# 786

# Aldy source: cn.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from __future__ import division
from builtins import zip
from builtins import range

import collections
import copy
import os

from . import lpinterface

from .common import *


def get_difference(coverage, unique_regions):
	if '7.pce' in coverage and '6.pce' not in coverage:
		coverage['6.pce'] = 0
	if '7.1e' in coverage:
		wo = [(coverage['6.' + r], coverage['7.' + r], coverage['6.' + r] - coverage['7.' + r]) for r in unique_regions]
	else:
		wo = [(coverage['6.' + r], 0, coverage['6.' + r]) for r in unique_regions]
	return dict(list(zip(unique_regions, wo)))


def load_profile(profile_path):
	result = collections.defaultdict(lambda: collections.defaultdict(int))
	with open(profile_path) as f:
		for line in f:
			if line[0] == '#':
				continue
			line = line.strip().split()[1:]
			if not line[0].startswith('chr'):
				line[0] = 'chr' + line[0]
			result[line[0]][int(line[1])] = float(line[2])
	return result


def normalize_coverage(gene, sam, profile):
	sam_ref = sum(sam.cnv_coverage[i] for i in range(gene.cnv_region[1], gene.cnv_region[2]))
	cnv_ref = sum(profile[gene.cnv_region[0]][i] for i in range(gene.cnv_region[1], gene.cnv_region[2]))

	cnv_ratio = float(cnv_ref) / sam_ref
	log.debug('CNV factor: {} ({})', cnv_ratio, 1.0 / cnv_ratio)

	sam.location_cnv = {}
	sam.region_coverage = {}
	for k, r in sorted(gene.regions.items()):
		s = sum(sam.unfiltered_coverage[i] for i in range(r[0], r[1] + 1))
		p = sum(profile[gene.region[0]][i] for i in range(r[0], r[1] + 1))
		sam.location_cnv.update({
			i: cnv_ratio * float(sam.unfiltered_coverage[i]) / profile[gene.region[0]][i]
			if profile[gene.region[0]][i] > 0 else 0.0
			for i in range(r[0], r[1] + 1)
		})
		sam.region_coverage[k] = (cnv_ratio * float(s) / p) if p != 0 else 0.0

def estimate_cn(gene, sam, profile, solution, solver):
	log.debug('== Copy Number Estimator ==')

	if solution is not None:
		solution = solution.split(',')
		solution = collections.Counter(solution)
		result = []
		res_cn = collections.defaultdict(int)
		log.debug('  Solutions (user-provided):')
		for sol in solution:
			if sol not in gene.cnv_configurations:
				raise Exception(
					('Given copy number solution contains unknown copy number configuration {}. ' + 
					 'Please run aldy --show-cn to list the valid configurations').format(sol)
				)
			for i in range(solution[sol]):
				result.append((sol, i))
				log.debug('    {}', (sol, i))
				for r in gene.cnv_configurations[sol]:
					res_cn[r] += gene.cnv_configurations[sol][r]
		return [(tuple(result), res_cn)]

	profile_path = script_path('aldy.resources.profiles', '{}.profile'.format(profile.lower()))
	if os.path.exists(profile_path):
		profile = load_profile(profile_path)
	else:
		profile = load_profile(profile)
	normalize_coverage(gene, sam, profile)

	coverage_diff = get_difference(sam.region_coverage, gene.unique_regions)

	log.debug('Coverage per region')
	log.debug('  {:5}: {} {}', 'region', ' 2D6', ' 2D7')

	for r in sorted(sam.region_coverage, key=region_sort_key):
		if r[0] != '6':
			continue
		if '7' + r[1:] in sam.region_coverage:
			log.debug(
				'  {:5}: {:5.2f} {:5.2f} {}', r[2:],
				sam.region_coverage['6' + r[1:]],
				sam.region_coverage['7' + r[1:]],
				'= {:5.2f}'.format(coverage_diff[r[2:]][2]) if r[2:] in coverage_diff else ''
			)
		else:
			log.debug('  {:5}: {:5.2f}', r[2:], sam.region_coverage['6' + r[1:]])

	structures = {(name, 0): structure for name, structure in gene.cnv_configurations.items()}

	# create special allele for any possible copy
	max_observed_cn = int(max(round(y) for r, y in sam.region_coverage.items())) + 1
	log.debug('Max. CN = {}', max_observed_cn)

	for a, _ in list(structures.keys()): # (a, -1) = a^ in old notation
		structures[(a, -1)] = structures[(a, 0)]
	for a, ai in list(structures.keys()):
		if ai == -1:
			continue
		if a == gene.deletion_allele:
			continue
		for i in range(1, max_observed_cn):
			structures[(a, i)] = copy.deepcopy(structures[(a, 0)])
			for r in gene.regions:
				if r.startswith('7.'):
					structures[(a, i)][r] = structures[(a, i)][r] - 1

	c = lpinterface.model('aldy_cnv', solver)

	# add one binary variable to model for any allele copy
	A = {(a, ai): c.addVar(vtype='B', name='A_{}_{}'.format(a, ai)) for a, ai in structures}

	pseudo_inducing = c.quicksum(A[a] for a in A if a[1] <= 0)
	c.addConstr(pseudo_inducing == 2)

	for a, ai in structures:
		if ai == 0 and (a, -1) in structures:
			log.trace('M  Sentinel contraint for {}: A_{}_-1 <= A_{}', a, a, a)
			c.addConstr(A[(a, -1)] <= A[(a, 0)])

	for a, ai in structures:
		if ai != 0:
			continue
		if a == gene.deletion_allele:
			continue
		for i in range(1, max_observed_cn):
			log.trace('M  Sentinel contraint for {}: A_{} <= A_{}', a, i, i - 1)
			c.addConstr(A[(a, i)] <= A[(a, i - 1)])

	E = {}
	for r, exp_cov in coverage_diff.items():
		exp_cov6, exp_cov7, exp_cov_diff = exp_cov
		expr = 0
		for a in structures:
			expr += A[a] * structures[a]['6.' + r]
			if '7.' + r in structures[a]:
				expr -= A[a] * structures[a]['7.' + r]
		E[r] = c.addVar(name=r, lb=-10, ub=10)

		log.trace('M  Contraint for {}: {} == E_{} + {}', r, exp_cov_diff, r, expr)
		c.addConstr(expr + E[r] == exp_cov_diff)

	objective = c.abssum(E.values(), coeffs={'pce': 1.5}) + .5 * sum(A.values())

	fusion_cost = c.quicksum(
		A[a] for a in A
		if a[0] in gene.fusions_left and a[0] != gene.deletion_allele
	)
	objective += .1 * fusion_cost
	log.trace('M  Objective: {}', objective)

	status, opt, solutions = c.solveAll(objective, A)
	log.debug('CN Solver status: {}, opt: {}', status, opt)

	log.debug('  Solutions:')
	for s in solutions:
		log.debug('    {}', s)
	for si, s in enumerate(solutions):
		res = collections.defaultdict(int)
		for sx in s:
			for r in structures[sx]:
				res[r] += structures[sx][r]
		solutions[si] = (s, res)

	return solutions



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# 786

# Aldy source: gene.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from __future__ import division
from __future__ import print_function
from builtins import next
from builtins import filter
from builtins import chr
from builtins import map
from builtins import range
from builtins import object
from collections import namedtuple as nt
from pprint import pprint

import os
import yaml
import collections

from .common import *


# disrupting_mutations, neutral_mutations
class Suballele(nt('Suballele', ['name', 'alt_names', 'neutral_mutations'])):
	"""
		alt_names			= list of other names for this suballele
		neutral_mutations 	= unordered set of Mutation
	"""
	def __repr__(self):
		return '{}: [{}]'.format('|'.join([self.name] + self.alt_names), ', '.join(map(str, self.neutral_mutations)))

	def __str__(self):
		return self.__repr__()


class Allele(nt('Allele', ['name', 'cnv_configuration', 'functional_mutations', 'suballeles'])):
	"""
		cnv_configuration 		= key for CNV configuration
		functional_mutations 	= unordered set of Mutation
		suballeles 				= dict of Suballele
	"""
	pass


class Mutation(nt('Mutation', ['pos', 'op', 'functional', 'aux'])):
	"""	Class representing the mutation
		pos 		= position in the reference genome
		op 			= mutation oprtation, one of the following:
			SNP.AB:	SNP from A to B (A, B are in [ACGT])
			INS.xx:	Insertion of xx (xx is [acgt]+)
			DEL.xx:	Deletion of xx (xx is [ACGT]+)
		functional  = does it impact the protein?
		aux 		= auxiliary information dict (right now dbSNP rsID and old notation)
	"""
	def __new__(self, pos, op, functional=0, aux=dict(), **kwargs):
		return super(Mutation, self).__new__(self, pos, op, functional, aux)

	def __repr__(self):
		return '{}:{}'.format(self.pos, self.op)

	def __str__(self):
		return self.__repr__()

	def __eq__(self, other):
		return (self.pos, self.op) == (other.pos, other.op)

	def __cmp__(self, other):
		return (self.pos, self.op).__cmp__((other.pos, other.op))

	def __hash__(self):
		return (self.pos, self.op).__hash__()

	def __ne__(self, other):
		return not self.__eq__(other)


class Gene(object):
	"""	Class representing the gene cluster structure
		name 		= gene name
		seq 		= gene nucleotide sequence
		region 		= (chr, start, end)
		cnv_region	= (chr, start, end)
		regions 	= { region_name: (start, end) }
			region_name is 'gene.name', where:
			gene is in {6, 7, ...}
			name is {1e, 1i, ..., pcr} etc.
		region_at 	= { pos: region_name }
		alleles 	= { allele_name: Allele }
		max_cn 		= { allele_name: max_cn }
		
		unique_regions	= [ region_name ... ]
		
		structures
		deletion_allele = name of deletion allele
		fusions 		= colection of fusions! {  }
		old_notation 	= lookup { Mutation: karolinska_snp_loci }
			this represents the canonical mutation code
			for the gene (e.g. -77:G>A)
	"""
	def __init__(self, path):
		self.load(path)

	def load(self, path):
		with open(path) as file:
			j = file.read()
			j = yaml.safe_load(j)

		## (1) Basics
		self.name = os.path.split(path)[-1].split('.')[0].upper()
		self.seq = j['seq']
		self.region = tuple(j['region'])
		self.cnv_region = tuple(j['cnv_region'])

		## (2) Set up gene regions
		# Convention: Main gene has prefix 6; pseudogenes are numbered 7, 8
		def calculate_regions(id, exons, special):
			regions = {'{}.{}e'.format(id, e): tuple(ex) for e, ex in exons.items()}
			for i in range(1, len(regions)):
				r1, r2 = regions['{}.{}e'.format(id, i + 1)], regions['{}.{}e'.format(id, i)]
				if r1 > r2:
					r1, r2 = r2, r1
				regions['{}.{}i'.format(id, i)] = (r1[1], r2[0])
			for n, r in special.items():
				regions['{}.{}'.format(id, n)] = tuple(r)
			return regions
		self.regions = calculate_regions(6, j['exons'], j['special_regions'])
		if 'pseudogenes' not in j:
			j['pseudogenes'] = {}
		for gi, g in enumerate(sorted(j['pseudogenes'])):
			g = j['pseudogenes'][g]
			self.regions.update(calculate_regions(7 + gi, g['exons'], g['special_regions']))
		self.region_at = collections.defaultdict(str, {
			i: n
			for n, j in self.regions.items()
			for i in range(j[0], j[1])
		})
		self.max_cn = j['max_cn']
		self.unique_regions = j['unique_regions']

		## (3) Set up alleles
		# is_left_fusion: 6.0 begin = 0
		self.common_tandems = []
		if 'common_tandems' in j:
			self.common_tandems = list(map(tuple, j['common_tandems']))
		alleles = {}
		fusions_left = {}
		fusions_right = {}
		self.descriptions = collections.defaultdict(str)
		self.descriptions['1'] = 'Normal allelic configuration: all regions present in both gene and pseudogene'
		self.old_notation = collections.defaultdict(str)
		self.deletion_allele = '0'
		for allele_name, allele in j['alleles'].items():
			allele_name = allele_name.split('*')[1]
			mutations = []
			if {'op': 'deletion'} in allele['mutations']:
				self.deletion_allele = allele_name
				self.descriptions[allele_name] = 'Gene deletion (only pseudogene is present)'
				mutations = []
			else:
				for m in allele['mutations']:
					# if 'pos' not in m:
						# continue
					if m['pos'] == 'pseudogene': # has only one mutation indicator
						if '-' in m['op']:
							fusions_left[allele_name] = m['op'][:-1][::-1] # eN -> Ne
							self.descriptions[allele_name] = "Fusion: pseudogene until {} followed by the gene".format(m['op'][:-1][::-1])
						else:
							if m['op'][-1] == '+':
								m['op'] = m['op'][:-1]
							fusions_right[allele_name] = m['op'][::-1]
							self.descriptions[allele_name] = "Conservation: Pseudogene retention after {} within the gene".format(m['op'][::-1])
					else:
						if 'old' in m:
							self.old_notation[Mutation(**m)] = m['old']
						m['aux'] = {'dbsnp': ','.join(m['dbsnp']), 'old': m['old']}
						mutations.append(Mutation(**m))
			alleles[allele_name] = Suballele(allele_name, [], mutations)

		## (4) Set up copy number structures
		regions = sorted(set(r[2:] for r in self.regions) - set('cnv'))
		fusions_left = { # pseudogene
			a: tuple(1 for r in regions if r < l) + (.5,) + tuple(0 for r in regions if r > l)
			for a, l in fusions_left.items()
		}
		fusions_left[self.deletion_allele] = tuple([1] * len(regions))
		fusions_left = {
			a: tuple(1 - y for y in l) + l
			for a, l in fusions_left.items()
		}
		fusions_right = { # pseudogene
			a: tuple(0 for r in regions if r < l) + tuple(1 for r in regions if r >= l)
			for a, l in fusions_right.items()
		}
		fusions_right = {
			a: tuple(1 - y for y in l) + tuple(1 + y for y in l)
			for a, l in fusions_right.items()
		}
		self.fusions = dict(list(fusions_left.items()) + list(fusions_right.items()))

		self.cnv_configurations = collections.defaultdict(set)
		for a, c in self.fusions.items():
			self.cnv_configurations[c].add(a)
		reference_cnv = tuple([1] * len(regions) * 2)
		self.cnv_configurations[reference_cnv] |= set(a for a in alleles if a not in self.fusions)
		self.cnv_configurations = {
			('1' if c == reference_cnv else allele_key(min(a))): (dict(
				[('6.' + regions[yi], y) for yi, y in enumerate(c[:len(c) // 2])] +
				[('7.' + regions[yi], y) for yi, y in enumerate(c[len(c) // 2:])]
			), a) for c, a in self.cnv_configurations.items()
		}
		self.descriptions = {allele_key(a): d for a, d in self.descriptions.items()}
		if self.name == 'CYP2D6': # This is necessary hack
			self.cnv_configurations['1'][0]['6.pce'] = 0
		self.fusions = { # TODO: check assumption that each fusion has unique allele_key
			allele_key(a): dict(
				[('6.' + regions[yi], y) for yi, y in enumerate(t[:len(t) // 2])] +
				[('7.' + regions[yi], y) for yi, y in enumerate(t[len(t) // 2:])]
			)
			for a, t in self.fusions.items()
		}
		self.fusions_left = {allele_key(a): c for a, c in fusions_left.items()}

		## (5) Set up major and minor allele structures
		self.alleles = collections.defaultdict(set)
		for a in alleles:
			cnv_configuration = next(cn for cn, (conf, alls) in self.cnv_configurations.items() if a in alls)
			self.alleles[tuple([cnv_configuration] + sorted(m for m in alleles[a].neutral_mutations if m.functional))].add(a)
		self.alleles = {
			min(suballeles): Allele('', key[0], key[1:], suballeles)
			for key, suballeles in self.alleles.items()
		}
		self.cnv_configurations = {cn: config[0] for cn, config in self.cnv_configurations.items()}
		used_names = {}
		for an, a in self.alleles.items():
			name = allele_key(an)
			if name in used_names: # append ' to the proteins with same numbers to distinguish them
				used_names[name] += 1
				name += '.' + chr(used_names[name] + ord('a'))
			else:
				used_names[name] = -1
			self.alleles[an] = Allele(
				name,
				a.cnv_configuration,
				set(a.functional_mutations),
				{sa: Suballele(sa, [], set(alleles[sa].neutral_mutations) - set(a.functional_mutations)) for sa in a.suballeles}
			)
		self.alleles = {a.name: a for a in self.alleles.values()}

		## (6) Add possible partial fusion products
		# TODO: currently only empty left fusions: extend to all fusions (CHECK does it happen)?
		def filter_missing(f, m):
			def filter_f(m):
				return self.fusions[f][self.region_at[m.pos]] > 0
			return set(filter(filter_f, m))
		for f in self.fusions_left:
			if f == self.deletion_allele:
				continue
			# do not extend left alleles which are already defined by some SNVs
			if len(self.alleles[f].functional_mutations) > 0:
				continue
			add = {}
			for an, a in sorted(list(self.alleles.items()), key=lambda x: sort_key(x[0])):
				if a.cnv_configuration != '1':
					continue
				n = f + '/' + an
				key = sorted_tuple(filter_missing(f, a.functional_mutations))
				if key in add:
					add[key] = Allele(
						add[key].name, add[key].cnv_configuration,
						add[key].functional_mutations,
						dict(
							list(add[key].suballeles.items()) +
							[(sa, Suballele(sa, [], filter_missing(f, set(alleles[sa].neutral_mutations) - set(a.functional_mutations)))) for sa in a.suballeles]
						)
					)
				else:
					add[key] = Allele(
						n, f,
						filter_missing(f, a.functional_mutations),
						{sa: Suballele(sa, [], filter_missing(f, set(alleles[sa].neutral_mutations) - set(a.functional_mutations))) for sa in a.suballeles}
					)
			del self.alleles[f]
			self.alleles.update({a.name: a for a in add.values()})
		for an, a in self.alleles.items():
			suballeles = collections.defaultdict(list)
			for s in a.suballeles:
				key = sorted_tuple(a.suballeles[s].neutral_mutations)
				suballeles[key].append(s)
			self.alleles[an] = Allele(
				self.alleles[an].name,
				self.alleles[an].cnv_configuration,
				self.alleles[an].functional_mutations,
				{min(sa): Suballele(min(sa), list(set(sa) - set([min(sa)])), set(nm)) for nm, sa in suballeles.items()}
			)

		## (7) Set up coding region structure for aminoacid calculation
		self.coding_region = {}
		self.rev_comp = True if j['rev_comp'] == 1 else False
		for ri, (start, end) in self.regions.items():
			p = re.match(r'6\.(\d+)e', ri)
			if not p:
				continue
			for i in range(start, end):
				self.coding_region[i] = self.seq[i - self.region[1]]

		seq = ''.join(self.coding_region[i] for i in sorted(self.coding_region.keys()))
		self.aminoacid = seq_to_amino(rev_comp(seq) if self.rev_comp else seq)

		self.mutations = {}
		for _, a in self.alleles.items():
			self.mutations.update({(m.pos, m.op): m for m in a.functional_mutations})
			for _, sa in a.suballeles.items():
				self.mutations.update({(m.pos, m.op): m for m in sa.neutral_mutations})

	def print_configurations(self):
		log.error('Gene {}\n', self.name)
		for name, config in sorted(self.cnv_configurations.items()):
			c6 = [c[2:] for c in config if c.startswith('6.')]
			c7 = [c[2:] for c in config if c.startswith('7.')]
			labels = sorted(list(set(c6) | set(c7)))
			max_width = str(max(3, max(len(x) for x in labels)) + 1)

			description = '{}\n'.format(self.descriptions[name]) if name in self.descriptions else ''
			alleles = sorted(set('*' + a.split('/')[0] for a in self.alleles if self.alleles[a].cnv_configuration == name))
			description += 'Alleles: ' 
			for i in range(0, len(alleles), 12):
				if i != 0: description += '\n         ';
				description += ', '.join(alleles[i:i + 12]) 

			log.warn('Name: {}\n{}'.format(name, description))
			print('            ', end='')
			for c in labels:
				print(('{:>' + max_width + '}').format(c), end='')
			print(' ')
			print('Gene:       ', end='')
			for c in labels:
				if '6.' + c in config:
					print(('{:' + max_width + '}').format(config['6.' + c]), end='')
			print(' ')
			print('Pseudogene: ', end='')
			for c in labels:
				if '7.' + c in config:
					print(('{:' + max_width + '}').format(config['7.' + c]), end='')
			print('\n')




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

exec(open('aldy/version.py').read())

setup(
    name='aldy',
    entry_points={
        'console_scripts': [
            'aldy = aldy.__main__:main'
        ]
    },
    packages=find_packages(),
    package_data={
        'aldy.resources': [
            '*.md', 'aldy/resources/*.md', 
            '*.bam', 'aldy/resources/*.bam',
            '*.bai', 'aldy/resources/*.bai'],
    	'aldy.resources.genes': ['*.yml', 'aldy/resources/genes/*.yml'],
    	'aldy.resources.profiles': ['*.profile', 'aldy/resources/profiles/*.profile']
    },
    install_requires=['pyyaml', 'logbook', 'six', 'pysam', 'future'],
    version=__version__,
    description='A tool for allelic decomposition and exact genotyping of highly polymorphic and structurally variant genes',
    author='Ibrahim Numanagić',
    author_email='inumanag@mit.edu',
    url='http://aldy.csail.mit.edu/',
    download_url='https://github.com/inumanag/aldy/tarball/master',
    keywords=['cyp2d6', 'adme', 'genotyping', 'illumina', 'pgrnseq', 'getrm', 'allele']
)



# 786

# Aldy source: lpinterface.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from builtins import object

import importlib

from .common import log


def model(name, solver):
	def test_gurobi(name):
		try:
			model = Gurobi(name)
			log.debug('Using Gurobi')
		except ImportError as e:
			log.warn('Gurobi not found. Please install Gurobi and gurobipy Python package.')
			log.error('{}', e)
			model = None
		return model

	# CPLEX does not support our quadratic models
	# def test_cplex(name):
	# 	try:
	# 		model = CPLEX(name)
	# 		log.warn('Using CPLEX')
	# 	except:
	# 		log.warn('CPLEX not found. Please install CPLEX and the following Python packages: cplex and docplex.')
	# 		model = None
	# 	return model

	def test_scip(name):
		try:
			model = SCIP(name)
			log.debug('Using SCIP')
		except ImportError as e:
			log.warn('SCIP not found. Please install SCIP and pyscipopt Python package.')
			log.error('{}', e)
			model = None
		return model

	if solver == 'any':
		model = test_gurobi(name)
		# if model is None:
		# 	model = test_cplex(name)
		if model is None:
			model = test_scip(name)
		if model is None:
			raise Exception('No IP solver found. Aldy cannot solve any problems without matching IP solver. Please try installing Gurobi or SCIP.')
		return model
	else:
		fname = 'test_' + solver
		if fname in locals():
			return locals()[fname](name)
		else:
			raise Exception('IP solver {} is not supported'.format(solver))


def get_all_solutions(c, var, opt, candidates, iteration=0, mem=[]):
	if candidates in mem:
		return []
	solutions = []
	for a in candidates:
		c.changeUb(var[a], 0)
		try:
			status, obj = c.solve()
			if status == 'optimal' and abs(obj - opt) < 1e-6:
				new_solution = set(a for a, y in var.items() if round(c.getValue(y)) > 0)
				new_candidates = (set(candidates) - set([a])) | new_solution
				solutions.append(frozenset(new_solution))
				solutions += get_all_solutions(c, var, opt, new_candidates, iteration + 1, mem)
		except NoSolutionsError:
			pass
		c.changeUb(var[a], 1)
		mem.append(candidates)
	return solutions


class NoSolutionsError(Exception):
	pass


class Gurobi(object):
	def __init__(self, name):
		self.gurobipy = importlib.import_module('gurobipy')

		self.INF = self.gurobipy.GRB.INFINITY
		self.GUROBI_STATUS = {
			getattr(self.gurobipy.GRB.status, v): v
			for v in dir(self.gurobipy.GRB.status)
			if v[:2] != '__'
		}
		self.model = self.gurobipy.Model(name)
		self.model.reset()

	def addConstr(self, *args, **kwargs):
		return self.model.addConstr(*args, **kwargs)

	def addVar(self, *args, **kwargs):
		if 'vtype' in kwargs and kwargs['vtype'] == 'B':
			kwargs['vtype'] = self.gurobipy.GRB.BINARY
		v = self.model.addVar(*args, **kwargs)
		self.model.update()
		return v

	def quicksum(self, expr):
		return self.gurobipy.quicksum(expr)

	def update(self):
		self.model.update()

	def varName(self, var):
		return var.varName

	def abssum(self, vars, coeffs=None):
		total = 0
		for v in vars:
			name = self.varName(v)
			coeff = 1 if coeffs is None or name not in coeffs else coeffs[name]
			absvar = self.addVar()
			total += absvar * coeff
			self.addConstr(absvar + v >= 0)
			self.addConstr(absvar - v >= 0)
		return total

	def solve(self, objective=None, method='min'): # ret obj val
		if objective is not None:
			self.objective = objective
		self.model.setObjective(
			self.objective,
			self.gurobipy.GRB.MINIMIZE if method == 'min' else self.gurobipy.GRB.MAXIMIZE
		)
		self.model.update()

		# self.model.params.timeLimit = 60
		# self.model.params.threads = 2
		self.model.params.outputFlag = 0
		self.model.params.logFile = ''
		self.model.optimize()

		status = self.GUROBI_STATUS[self.model.status]
		if self.model.status == self.gurobipy.GRB.INFEASIBLE:
			raise NoSolutionsError(status)
		return status.lower(), self.model.objVal

	def solveAll(self, objective, var, method='min'):
		status, opt_value = self.solve(objective, method)
		solutions = [frozenset(a for a, y in var.items() if round(self.getValue(y)) > 0)]
		solutions += get_all_solutions(
			self, var, opt_value,
			candidates=solutions[0], iteration=0, mem=list()
		)
		solutions = [tuple(sorted(y for y in x)) for x in solutions]

		solutions = list(set(solutions))
		return status, opt_value, solutions

	def getValue(self, var):
		return var.x

	def changeUb(self, var, ub):
		var.ub = ub
		self.model.update()


class SCIP(Gurobi):
	def __init__(self, name):
		self.pyscipopt = importlib.import_module('pyscipopt')
		self.INF = 1e20
		self.model = self.pyscipopt.Model(name)

	def update(self):
		pass

	def addConstr(self, *args, **kwargs):
		return self.model.addCons(*args, **kwargs)

	def addVar(self, *args, **kwargs):
		return self.model.addVar(*args, **kwargs)

	def quicksum(self, expr):
		return self.pyscipopt.quicksum(expr)

	def solve(self, objective=None, method='min'):
		if objective is not None:
			self.objective = objective
		self.model.setObjective(
			self.objective,
			'minimize' if method == 'min' else 'maximize'
		)
		# self.model.params.timeLimit = 60
		self.model.setRealParam('limits/time', 120)
		self.model.hideOutput()
		self.model.optimize()

		status = self.model.getStatus()
		if status == 'infeasible':
			raise NoSolutionsError(status)
		return status, self.model.getObjVal()

	def varName(self, var):
		return var.name

	def getValue(self, var):
		return self.model.getVal(var)

	def changeUb(self, var, ub):
		self.model.freeTransform()
		self.model.chgVarUb(var, ub)


# CPLEX does not support our quadratic models.

# class CPLEX:
# 	def __init__(self, name):
# 		self.docplex = importlib.import_module('docplex.mp.model')
# 		self.model = self.docplex.Model(name=name)
# 		self.INF = self.model.infinity
# 		print self.INF

# 	def addConstr(self, *args, **kwargs):
# 		return self.model.add_constraint(*args, **kwargs)

# 	def addVar(self, *args, **kwargs):
# 		vtype = 'c'
# 		if 'vtype' in kwargs:
# 			vtype = 'b' if kwargs['vtype'] == 'B' else 'c'
# 			del kwargs['vtype']
# 		if vtype == 'b':
# 			v = self.model.binary_var(*args, **kwargs)
# 		else:
# 			v = self.model.continuous_var(*args, **kwargs)
# 		return v

# 	def quicksum(self, expr):
# 		return self.model.sum(expr)

# 	def update(self):
# 		pass

# 	def varName(self, var):
# 		return var.varName

# 	def abssum(self, vars, coeffs=None):
# 		return self.quicksum([self.model.abs(v) for v in vars])
		
# 	def solve(self, objective=None, method='min'): # ret obj val
# 		if objective is not None:
# 			self.objective = objective
# 		if method == 'min':
# 			self.model.minimize(self.objective)
# 		else:
# 			self.model.maximize(self.objective)
		
# 		# self.model.params.timeLimit = 60
# 		# self.model.params.threads = 2
# 		# self.model.params.outputFlag = 0
# 		# self.model.params.logFile = ''
		
# 		self.model.export_as_lp(path='./MODEL_DBG.LP')
# 		params = {}
# 		try:
# 			self.solution = self.model.solve(cplex_parameters=params, agent='local', log_output=False)
# 		except:
# 			self.model.export_as_lp(path='./MODEL_DBG.LP')
# 			self.solution = None
# 			exit(1)
# 		if not self.solution:
# 			raise NoSolutionsError(self.solution)
# 		return 'optimal', self.solution.get_objective_value()

# 	def solveAll(self, objective, var, method='min'):
# 		status, opt_value = self.solve(objective, method)
# 		solutions = [frozenset(a for a, y in var.iteritems() if round(self.getValue(y)) > 0)]
# 		solutions += get_all_solutions(
# 			self, var, opt_value,
# 			candidates=solutions[0], iteration=0, mem=list()
# 		)
# 		solutions = map(lambda x: tuple(sorted(y for y in x)), solutions)

# 		solutions = list(set(solutions))
# 		return status, opt_value, solutions

# 	def getValue(self, var):
# 		return self.solution.get_value(var)

# 	def changeUb(self, var, ub):
# 		var.ub = ub



#!/usr/bin/env python
# 786

# Aldy source: test.py
#   This file is subject to the terms and conditions defined in
#   file 'LICENSE', which is part of this source code package.


from __future__ import print_function
from builtins import str

import re
import os
import sys
import pandas as pd
import logbook
import logbook.more

from . import genotype
from . import solver

from .gene import SAM
from .common import log, LOG_FORMAT, colorize, allele_key


def sortkey(x):
	return int(allele_key(x))


def test_single(sample, location, expected, profile, threshold, gene):
	expected = [r for r in expected if not pd.isnull(r)]
	message = '{} - {}'.format(sample, ' or '.join(expected))
	message = colorize('{:30}'.format(message, 'teal'))
	if '' not in expected:
		expected = [[str(x).strip() for x in re.split('[/\+]', r)] for r in expected]
		expected = set(tuple(sorted(r, key=sortkey)) for r in expected)
	else:
		expected = set()

	try:
		solutions = genotype.genotype(location, gene, profile, threshold)
	except:
		logbook.error('{} {}', message, colorize('CRASH ', 'red'))
		exit(1)

	def fix(s):
		return re.split('(\d+)', s)[1]
	orig_solutions = solutions
	solutions = set(tuple(sorted((fix(p) for p in s), key=sortkey)) for s in solutions)
	expected = set(tuple(sorted((fix(p) for p in s), key=sortkey)) for s in expected)

	if solutions == expected:
		logbook.info('{} {} {}', message, colorize('OK   ', 'green'), list(orig_solutions))
		return 1
	elif solutions <= expected and len(solutions) != 0:
		logbook.info('{} {} {}', message, colorize('OK<  ', 'green'), list(orig_solutions))
		return 2
	elif len(expected & solutions) > 0:
		logbook.warn('{} {} {}', message, colorize('MULTI', 'yellow'), list(orig_solutions))
		return 3
	else:
		logbook.error('{} {} {}', message, colorize('FAIL ', 'red'), list(orig_solutions))
		return 0


def test_samples(samples, location, profile, threshold, gene='cyp2d6'):
	from pathos.multiprocessing import ProcessingPool as Pool
	
	pool = Pool(processes=16)
	
	def f(s):
		global counter
		sample, expected = s[0], s[1:]
		# if sample not in T: continue
		loc = location.format(sample)
		if not os.path.exists(loc): # or sample == 'PGXT122':
			return 0
		try:
			res = test_single(sample, loc, expected, profile, threshold, gene)
		except Exception as e:
			print(s, e)
			res = 0
		# sys.stderr.write('-')
		return res
	
	# sys.stderr.write(' ' * len(samples) + '|\r')
	result = pool.map(f, samples)
	logbook.warn(
		'Passed {} out of {} ({} subset, {} multi)\n',
		colorize(str(len([x for x in result if x > 0])), 'green'),
		colorize(str(len(result)), 'blue'),
		colorize(str(len([x for x in result if x == 2])), 'yellow'),
		colorize(str(len([x for x in result if x == 3])), 'yellow')
	)


if __name__ == "__main__":
	path = os.path.dirname(os.path.realpath(__file__))
	
	if '-f' in sys.argv:
		os.system(
			'curl -L "https://docs.google.com/spreadsheets/d/17GUAk5CP9rW7Dj918biIY5bKyCHY_45iq9iKKqX8yP0/export?format=xlsx"' +
			' > {}/../resources/tests.xlsx'.format(path))

	sh = logbook.more.ColorizedStderrHandler(format_string=LOG_FORMAT, level='WARNING')
	sh.push_application()
	log_file = path + '/../test-results.txt'
	fh = logbook.FileHandler(log_file, format_string=LOG_FORMAT, mode='w', bubble=True, level='INFO')
	fh.push_application()

	SAM.CACHE = True
	solver.SHOW_CN_INFO = False

	def get_samples(sheet, gene, samples_path, profile, threshold):
		if 'CDC' in sheet:
			data = pd.read_excel(path + '/../resources/tests.xlsx', header=[0, 2], sheetname=sheet, parse_dates=False)
			data = data.set_index(['PGXT'])
		else:
			data = pd.read_excel(path + '/../resources/tests.xlsx', header=[0, 1], sheetname=sheet, parse_dates=False)
		samples = []
		for r in data[gene].iterrows():
			sample_id = r[0]
			if isinstance(sample_id, tuple):
				sample_id = sample_id[0].upper()
			p = [sample_id, r[1].Validated]
			if not pd.isnull(r[1].Additional):
				p += r[1].Additional.split(';')
			samples.append(p)

		log.warn('{} [{}]', sheet, gene)
		test_samples(samples, path + samples_path, profile=profile, threshold=threshold, gene=gene.lower())

	# get_samples('PGRNseq-v2 (Old)', gene='CYP2D6', samples_path='/../data/pgrnseq-old/{}.dz', profile='pgrnseq-v2', threshold=.5)
	# get_samples('PGRNseq-v2',       gene='CYP2D6', samples_path='/../data/pgrnseq-96/{}.dz',  profile='pgrnseq-v2', threshold=.5)
	
	# get_samples('Illumina',      gene='CYP2D6', samples_path='/../data/illumina/{}.bam',        profile='illumina', threshold=.5)
	# get_samples('Illumina (IU)', gene='CYP2D6', samples_path='/../data/illumina-milan/{}.bam',  profile='illumina', threshold=.5)
	
	# get_samples('PGRNseq-v1 (CDC)', gene='CYP2D6',  samples_path='/../data/pgrnseq-cdc/{}.dz',  profile='pgrnseq-v1', threshold=.5)
	# get_samples('PGRNseq-v1 (CDC)', gene='CYP2A6',  samples_path='/../data/pgrnseq-cdc/{}.dz',  profile='pgrnseq-v1', threshold=.5)
	get_samples('PGRNseq-v1 (CDC)', gene='CYP2C19', samples_path='/../data/pgrnseq-cdc/{}.dz', profile='pgrnseq-v1', threshold=.5)
	get_samples('PGRNseq-v1 (CDC)', gene='CYP2C8', samples_path='/../data/pgrnseq-cdc/{}.dz', profile='pgrnseq-v1', threshold=.5)
	get_samples('PGRNseq-v1 (CDC)', gene='CYP2C9', samples_path='/../data/pgrnseq-cdc/{}.dz', profile='pgrnseq-v1', threshold=.5)
	get_samples('PGRNseq-v1 (CDC)', gene='CYP3A4', samples_path='/../data/pgrnseq-cdc/{}.dz', profile='pgrnseq-v1', threshold=.5)
	get_samples('PGRNseq-v1 (CDC)', gene='CYP3A5', samples_path='/../data/pgrnseq-cdc/{}.dz', profile='pgrnseq-v1', threshold=.5)
	get_samples('PGRNseq-v1 (CDC)', gene='CYP4F2', samples_path='/../data/pgrnseq-cdc/{}.dz', profile='pgrnseq-v1', threshold=.5)
	get_samples('PGRNseq-v1 (CDC)', gene='TPMT', samples_path='/../data/pgrnseq-cdc/{}.dz', profile='pgrnseq-v1', threshold=.5)
	get_samples('PGRNseq-v1 (CDC)', gene='DPYD', samples_path='/../data/pgrnseq-cdc/{}.dz', profile='pgrnseq-v1', threshold=.5)
	
	# TODO:
	# get_samples('PGRNseq-v1 (CDC)', gene='CYP2B6',  samples_path='/../data/pgrnseq-cdc/{}.dz',  profile='pgrnseq-v1', threshold=.5)
	# get_samples('PGRNseq-v1 (CDC)', gene='SLCO1B1', samples_path='/../data/pgrnseq-cdc/{}.dz',  profile='pgrnseq-v1', threshold=.5)
